//ModbusRTU.cpp

#include <Arduino.h>
#include "config.h"
#include "ModbusRTU.h"
#include "maindat.h"

DAT_MRTU  dmrtu;
uint8_t wcnt;	//количество прочитанных байтов

//инициализация
int8_t  sMRTU_INIT(void)
{
  memset(&dmrtu.rxbuf[0], 0, _szbuf);
  memset(&dmrtu.txbuf[0], 0, _szbuf);
  dmrtu.adrx = 0;
  dmrtu.adtx = 0;
  return 0;
}

/***************************/

//контрольная сумма
inline uint16_t sMRTU_CRC(uint8_t* data, uint16_t len)
{
#if(_CRC_USE_TABLES == 0)
	uint16_t res_CRC = 0xFFFF;
	uint16_t count = 0;
	uint8_t count_crc;
	uint8_t dt;
	while(count < len)
	{
		//
		count_crc = 0;
		dt = (uint8_t)(data[count]);
		res_CRC ^= (uint16_t)(dt);
		//
		while(count_crc < 8)
		{
			if((res_CRC & 0x0001) < 1)
			{
				res_CRC = (res_CRC >> 1) & 0x7FFF;
			}
			else
			{
				res_CRC = (res_CRC >> 1) & 0x7FFF;
				res_CRC ^= 0xA001;
			};
			count_crc++;
		};
		count++;
	}
  	return (res_CRC);
#elif(_CRC_USE_TABLES == 1)
	uint8_t uchCRCHi = 0xFF;
	uint8_t uchCRCLo = 0xFF;
	uint16_t uIndex ;
	while(len--)
	{
		uIndex = uchCRCHi ^ *data++ ;
		uchCRCHi = uchCRCLo ^ auchCRCHi[uIndex] ;
		uchCRCLo = auchCRCLo[uIndex] ;
	}
	//return (uint16_t)(uchCRCHi | uchCRCLo << 8) ;
	return (uint16_t)(uchCRCHi | uchCRCLo * 256) ;
#endif
}

/*
****************************
CMD3 - чтение двоичного
содержания регистров
****************************
*/
inline int8_t	MRTU_CMD3(void)
{
	uint16_t  crc; 			//подсчитанная контрольная сумма
	uint16_t  incrc;        	//принятая контрольная сумма
	uint16_t  sreg;			//стартовый регистр
	uint16_t  cntreg;		//количество регистров
	uint8_t   bcntreg;
	uint8_t	  cntbytes;		//счётчик байт в поле данных
	uint8_t	  bsreg;
	uint8_t   wregcnt;
	//стартовый регистр
	#if(_opt_8 == 0)
		sreg	= (uint16_t)(dmrtu.rxbuf[_ADDR_03_START_REG_H] * 256)
      			| (uint16_t)(dmrtu.rxbuf[_ADDR_03_START_REG_L]);
	#else
		bsreg	= dmrtu.rxbuf[_ADDR_03_START_REG_L];
	#endif
	//количество регистров
	#if(_opt_8 == 0)
		cntreg  = (uint16_t)(dmrtu.rxbuf[_ADDR_03_CNT_REG_H] * 256)
      			| (uint16_t)(dmrtu.rxbuf[_ADDR_03_CNT_REG_L]);
	#else
		bcntreg	= dmrtu.rxbuf[_ADDR_03_CNT_REG_L];
	#endif
	//чтение контрольной суммы
	incrc	= (uint16_t)(dmrtu.rxbuf[_ADDR_03_INCRC_H] * 256)
		| (uint16_t)(dmrtu.rxbuf[_ADDR_03_INCRC_L]);
	crc = sMRTU_CRC(&dmrtu.rxbuf[0], 6);
	if(crc == incrc)
      	{
        	/*
        	****************************
        	проверка границ данных
        	****************************
        	*/
		#if(_opt_8 == 0)
			bsreg	= (uint8_t)(sreg);
			bcntreg	= (uint8_t)(cntreg);
		#endif
		//проверка предельного размера
		if(bcntreg >= _cntreg)
		{
			dmrtu.txbuf[0] = _MRTU_Addr;
			dmrtu.txbuf[1] = 3 | _CMD_ERR_MODE;
			dmrtu.txbuf[2] = _ILLEGAL_DATA_ADDRESS;
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 3);
			dmrtu.txbuf[3] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[4] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 5);
		}
		//ComMotion
		else if((bsreg >= _ComMotion_RMS) && 
			(bsreg < (_ComMotion_RMS + _ComMotion_SzRM)) && 
			((bsreg + bcntreg) <= (_ComMotion_RMS + _ComMotion_SzRM)))
		{
			dmrtu.txbuf[_ADDR_MADDR]  	= _MRTU_Addr;			//адрес устройства
			dmrtu.txbuf[_ADDR_MCMD]  	= 3;				//команда
			dmrtu.txbuf[_ADDR_03_CNT_BYTES]	= bcntreg * 2;			//счётчик байтов для передачи
			//
			bsreg 				-= _ComMotion_RMS;		//стартовый регистр
			cntbytes			= _ADDR_03_DATA;		//счётчик байтов + смещение
			bcntreg				+= bsreg;			//количество регистров со смещением
			wregcnt				= bsreg;			//счётчик регистров со смещением
			while(wregcnt < bcntreg)
			{	
				dmrtu.txbuf[cntbytes] = (uint8_t)(dmotor[wregcnt] >> 8);
				dmrtu.txbuf[cntbytes + 1] = (uint8_t)(dmotor[wregcnt] & 0xFF);
				cntbytes	+= 2;
				wregcnt++;
			}
			crc = sMRTU_CRC(&dmrtu.txbuf[0], cntbytes);
			dmrtu.txbuf[cntbytes] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[cntbytes + 1] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], cntbytes + 2);
		}
		//Multiservo
		else if((bsreg >= _Multiservo_RSS1) && 
			(bsreg < (_Multiservo_RSS1 + _Multiservo_SzRS)) && 
			((bsreg + bcntreg) <= (_Multiservo_RSS1 + _Multiservo_SzRS)))
		{
			dmrtu.txbuf[_ADDR_MADDR]  	=  _MRTU_Addr;			//адрес устройства
			dmrtu.txbuf[_ADDR_MCMD]  	=  3;				//команда
			dmrtu.txbuf[_ADDR_03_CNT_BYTES]	= bcntreg * 2;			//счётчик байтов для передачи
			//
			bsreg 				-= _Multiservo_RSS1;		//стартовый регистр
			cntbytes			= _ADDR_03_DATA;		//счётчик байтов + смещение
			bcntreg				+= bsreg;			//количество регистров со смещением
			wregcnt				= bsreg;			//счётчик регистров со смещением
			while(wregcnt < bcntreg)
			{	
				dmrtu.txbuf[cntbytes] = (uint8_t)(dmotor[wregcnt] >> 8);
				dmrtu.txbuf[cntbytes + 1] = (uint8_t)(dmotor[wregcnt] & 0xFF);
				cntbytes	+= 2;
				wregcnt++;
			}
			crc = sMRTU_CRC(&dmrtu.txbuf[0], cntbytes);
			dmrtu.txbuf[cntbytes] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[cntbytes + 1] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], cntbytes + 2);
		}
		else
		{
			dmrtu.txbuf[0] = _MRTU_Addr;
			dmrtu.txbuf[1] = 3 | _CMD_ERR_MODE;
			dmrtu.txbuf[2] = _ILLEGAL_DATA_ADDRESS;
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 3);
			dmrtu.txbuf[3] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[4] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 5);
		}
	}
	return 0;
}

/*
****************************
CMD6 - запись в один регистр
****************************
*/
inline int8_t	MRTU_CMD6(void)
{
	uint16_t  crc; 			//подсчитанная контрольная сумма
	uint16_t  incrc;        	//принятая контрольная сумма
	uint16_t  sreg;			//стартовый регистр
	uint8_t	  bsreg;
	uint16_t  datreg;
      	//стартовый регистр
	sreg	= (uint16_t)(dmrtu.rxbuf[_ADDR_06_ADDR_REG_H] * 256)
      		| (uint16_t)(dmrtu.rxbuf[_ADDR_06_ADDR_REG_L]);
      	//данные регистра
	datreg  = (uint16_t)(dmrtu.rxbuf[_ADDR_06_DAT_H] * 256)
      		| (uint16_t)(dmrtu.rxbuf[_ADDR_06_DAT_L]);
	//чтение контрольной суммы
	incrc	= (uint16_t)(dmrtu.rxbuf[_ADDR_06_INCRC_H] * 256)
		| (uint16_t)(dmrtu.rxbuf[_ADDR_06_INCRC_L]);
      	crc = sMRTU_CRC(&dmrtu.rxbuf[0], 6);
      	if(crc == incrc)
      	{
        	/*
        	****************************
        	проверка границ данных
        	****************************
        	*/
		bsreg = (uint8_t)(sreg);
        	//ComMotion
        	if((bsreg >= _ComMotion_RMS) && (bsreg < (_ComMotion_RMS + _ComMotion_SzRM)))
        	{
           		dmotor[sreg - _ComMotion_RMS] = (int16_t)(datreg);
           		memcpy(&dmrtu.txbuf[0], &dmrtu.rxbuf[0], 8);
           		Serial.write(&dmrtu.txbuf[0], 8);
        	}
		//Multiservo
		else if((bsreg >= _Multiservo_RSS1) && (bsreg < (_Multiservo_RSS1 + _Multiservo_SzRS)))
		{
			dservo[sreg - _Multiservo_RSS1] = (int16_t)(datreg);
			memcpy(&dmrtu.txbuf[0], &dmrtu.rxbuf[0], 8);
			Serial.write(&dmrtu.txbuf[0], 8);
		}
		// ILLEGAL DATA ADDRESS - вышли за предел данных
		else
		{
			dmrtu.txbuf[0] = _MRTU_Addr;
			dmrtu.txbuf[1] = 6 | _CMD_ERR_MODE;
			dmrtu.txbuf[2] = _ILLEGAL_DATA_ADDRESS;
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 3);
			dmrtu.txbuf[3] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[4] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 5);
		}
        //
	}
	return 0;
}

/*
****************************
CMD16 - запись в несколько 
регистров
****************************
*/
inline int8_t	MRTU_CMD16(void)
{
	uint16_t  crc; 			//подсчитанная контрольная сумма
	uint16_t  incrc;        	//принятая контрольная сумма
	#if(_opt_8 == 0)
		uint16_t  sreg;		//стартовый регистр
		uint16_t  cntreg;	//количество регистров
	#endif
	uint8_t	  bsreg;
	uint8_t   bcntreg;
	uint8_t	  cntbytes;		//счётчик байт в поле данных
	uint8_t   wregcnt;
	uint8_t   wcntb		= 0;
	//стартовый регистр
	#if(_opt_8 == 0)
		sreg  		= (uint16_t) (dmrtu.rxbuf[_ADDR_16_ADDR_REG_H] * 256)
				| (uint16_t) (dmrtu.rxbuf[_ADDR_16_ADDR_REG_L]);
	#else
		bsreg		= dmrtu.rxbuf[_ADDR_16_ADDR_REG_L];
	#endif
	//количество регистров
	#if(_opt_8 == 0)
		cntreg  	= (uint16_t) (dmrtu.rxbuf[_ADDR_16_CNT_REG_H] * 256)
				| (uint16_t) (dmrtu.rxbuf[_ADDR_16_CNT_REG_L]);
	#else
		bcntreg		= dmrtu.rxbuf[_ADDR_16_CNT_REG_L];
	#endif
	//количество байт данных
	cntbytes = dmrtu.rxbuf[_ADDR_16_CNT_BYTES];
	//чтение контрольной суммы
	incrc	= (uint16_t)(dmrtu.rxbuf[cntbytes + 7] * 256) | (uint16_t)(dmrtu.rxbuf[cntbytes + 8]);
	//подсчёт контрольной суммы
	crc = sMRTU_CRC(&dmrtu.rxbuf[0], cntbytes + 7);
	//
	if(crc == incrc)
	{
		/*
        	****************************
        	проверка границ данных
        	****************************
        	*/
		#if(_opt_8 == 0)
			bsreg	= (uint8_t)(sreg);
			bcntreg	= (uint8_t)(cntreg);
		#endif
		//проверка предельного размера
		if(bcntreg >= _cntreg)
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16 | _CMD_ERR_MODE;
			dmrtu.txbuf[2]  =  _ILLEGAL_DATA_ADDRESS;
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 3);
			dmrtu.txbuf[3] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[4] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 5);
		}
        	//ComMotion
		else if((bsreg >= _ComMotion_RMS) && 
			(bsreg < (_ComMotion_RMS + _ComMotion_SzRM)) && 
			((bsreg + bcntreg) <= (_ComMotion_RMS + _ComMotion_SzRM)))
		{
			bsreg 	-= _ComMotion_RMS;	//стартовый регистр
			wcntb 	+= _ADDR_16_DAT_H;	//счётчик байтов + смещение
			bcntreg	+= bsreg;		//количество регистров со смещением
			wregcnt	= bsreg;		//счётчик регистров со смещением
			while(wregcnt < bcntreg)
			{
				dmotor[wregcnt]	= ((int16_t)(dmrtu.rxbuf[wcntb]) * 256)
						| (int16_t)(dmrtu.rxbuf[wcntb + 1]);
				wregcnt++;
				wcntb += 2; 
			}
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16;
			dmrtu.txbuf[2]  =  dmrtu.rxbuf[2];
			dmrtu.txbuf[3]  =  dmrtu.rxbuf[3];
			dmrtu.txbuf[4]  =  dmrtu.rxbuf[4];
			dmrtu.txbuf[5]  =  dmrtu.rxbuf[5];
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 6);
			dmrtu.txbuf[6] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[7] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 8);
		}
		//Multiservo
		else if((bsreg >= _Multiservo_RSS1) && 
			(bsreg < (_Multiservo_RSS1 + _Multiservo_SzRS)) && 
			((bsreg + bcntreg) <= (_Multiservo_RSS1 + _Multiservo_SzRS)))
		{
			bsreg 	-= _Multiservo_RSS1;	//стартовый регистр
			wcntb 	+= _ADDR_16_DAT_H;	//счётчик байтов + смещение
			bcntreg	+= bsreg;		//количество регистров со смещением
			wregcnt	= bsreg;		//счётчик регистров со смещением
			while(wregcnt < bcntreg)
			{
				dservo[wregcnt]	= ((int16_t)(dmrtu.rxbuf[wcntb]) * 256)
						| (int16_t)(dmrtu.rxbuf[wcntb + 1]);
				wregcnt++;
				wcntb += 2;
			}
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16;
			dmrtu.txbuf[2]  =  dmrtu.rxbuf[2];
			dmrtu.txbuf[3]  =  dmrtu.rxbuf[3];
			dmrtu.txbuf[4]  =  dmrtu.rxbuf[4];
			dmrtu.txbuf[5]  =  dmrtu.rxbuf[5];
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 6);
			dmrtu.txbuf[6] = (uint8_t)(crc >> 8);
			dmrtu.txbuf[7] = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 8);
		}
		// ILLEGAL DATA ADDRESS - вышли за предел данных
		else
		{
			dmrtu.txbuf[0]  =  _MRTU_Addr;
			dmrtu.txbuf[1]  =  16 | _CMD_ERR_MODE;
			dmrtu.txbuf[2]  =  _ILLEGAL_DATA_ADDRESS;
			crc = sMRTU_CRC(&dmrtu.txbuf[0], 3);
			dmrtu.txbuf[3]  = (uint8_t)(crc >> 8);
			dmrtu.txbuf[4]  = (uint8_t)(crc & 0xFF);
			Serial.write(&dmrtu.txbuf[0], 5);
		}
	}
	return 0;
}

/*
****************************
неопределённая команда
****************************
*/
int8_t	MRTU_CMDunknown(void)
{
	uint16_t  crc; //подсчитанная контрольная сумма
	dmrtu.txbuf[0]  =  _MRTU_Addr;
       	dmrtu.txbuf[1]  =  dmrtu.rxbuf[1] | _CMD_ERR_MODE;
       	dmrtu.txbuf[2]	=  _ILLEGAL_FUNCTION;
       	crc = sMRTU_CRC(&dmrtu.txbuf[0], 3);
	dmrtu.txbuf[3] = (uint8_t)(crc >> 8);
	dmrtu.txbuf[4] = (uint8_t)(crc & 0xFF);
       	Serial.write(&dmrtu.txbuf[0], 5);
	return 0;
}

/*
****************************
приём команды и обработка
****************************
*/
int8_t  MRTU_rWORK(void)
{
	int8_t rcvsz = Serial.available();	//принятый размер
	//нет пакета для обработки
	if(rcvsz < 8)
		return 0;
	wcnt = Serial.readBytes((char*)(&dmrtu.rxbuf[0]), _szbuf);
	//не наш адрес
  	if(dmrtu.rxbuf[_ADDR_MADDR] != _MRTU_Addr)
		return 0;
	//определение команд
	switch(dmrtu.rxbuf[_ADDR_MCMD])
	{
		case 3:
			MRTU_CMD3();
			break;
		case 6:
			MRTU_CMD6();
			break;
		case 16:
			MRTU_CMD16();
			break;
		default:
			MRTU_CMDunknown();
			break;
	}
	Serial.flush();
	return 0;
}


